import bpy
import io
import xml.etree.ElementTree as ET
from contextlib import redirect_stdout, redirect_stderr

bl_info = {
	"name": "Import BeamNG.drive levels",
	"description": "Import levels from BeamNG.drive",
	"author": "DGIorio",
	"version": (1, 0),
	"blender": (2, 80, 0),
	"location": "File > Import > BeamNG.drive Level Importer",
	"warning": "",
	"wiki_url": "",
	"tracker_url": "",
	"support": "COMMUNITY",
	"category": "Import-Export"}

if bpy.app.version >= (2, 80, 0):
	bl_info["blender"] = (2, 80, 0)
else:
	bl_info["blender"] = (2, 79, 0)

#  Import levels from BeamNG.drive
#
#  This program is free software; you can redistribute it and/or
#  modify it.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

from bpy_extras.io_utils import ImportHelper, unpack_list, unpack_face_list
from bpy.props import StringProperty, BoolProperty, EnumProperty
from bpy.types import Operator
import math
import os

def read_mis(misPath, edit_texture_paths):
	srcPath = os.path.dirname(misPath)
	basename = os.path.basename(srcPath)
	
	## Read mis file
	with open(misPath, "r") as f:
		lines = f.readlines()
		for i, line in enumerate(lines):
			if "new TSStatic()" in line:
				shapeName = lines[i+1].split(" = ")[1].split(";")[0].replace('"', '')
				shapeName = shapeName.replace("levels/", "").replace("GridMap_v0422/", "").replace("/", "\\")	# ARRUMAR, nome tá fixo
				shapeName = os.path.join(srcPath, shapeName)
				
				position = lines[i+12].split(" = ")[1].replace(";", "").replace('"', '')
				position = [float(item) for item in position.split()]
				
				rotation = lines[i+13].split(" = ")[1].replace(";", "").replace('"', '')
				rotation = [float(item) for item in rotation.split()]
				
				scale = lines[i+14].split(" = ")[1].replace(";", "").replace('"', '')
				scale = [float(item) for item in scale.split()]
				
				if edit_texture_paths == True:
					# Editing texture paths
					resave = False
					tree = ET.parse(shapeName)
					tree.write(shapeName + ".bak")
					root = tree.getroot()
					prefix = root.tag.split("}")[0] + "}"
					#elems = root.findall('.//%simage' % prefix)
					elems = root.findall('.//%simage/%sinit_from' % (prefix,prefix))
					for elem in elems:
						#print("elem.attrib:", elem.attrib)
						#print("elem.text:", elem.text)
						tex_name = elem.text.split("/")[-1]
						tex_path = os.path.join(os.path.dirname(shapeName), tex_name)
						textures_folder = os.path.join(srcPath, "textures")
						textures_path = os.path.join(textures_folder, tex_name)
						# Checking if the texture is in the same folder as de collada file
						if os.path.isfile(tex_path):
							# Cheking if its path was already right
							if not os.path.isfile(os.path.join(os.path.dirname(shapeName), elem.text)):
								elem.text = tex_name
								resave = True
						# Checking if the texture is in the textures folder
						elif os.path.isfile(textures_path):
							elem.text = "/" + textures_path.replace("\\", "/")	# Collada path standard
							resave = True
						else:
							print("Texture not found", elem.text.split("/")[-1])
					if resave == True:
						tree.write(shapeName)
					else:
						os.remove(shapeName + ".bak")
				
				print("Importing", os.path.basename(shapeName))
				
				stdout = io.StringIO()
				with redirect_stdout(stdout), redirect_stderr(stdout):
					bpy.ops.wm.collada_import(filepath=shapeName)
				
				for obj in bpy.context.selected_objects:
					if obj.type == 'MESH' and 'Colmesh' in obj.name:
						bpy.data.objects.remove(obj, do_unlink=True)
						continue
					if obj.type == 'MESH' and 'Untitled' in obj.name:
						bpy.data.objects.remove(obj, do_unlink=True)
						continue
					if obj.type == 'EMPTY':
						bpy.data.objects.remove(obj, do_unlink=True)
						continue
					if obj.type == 'MESH':
						new_name = os.path.basename(shapeName).split(".")[0]
						if obj.data and obj.data.users == 1:
							obj.data.name = new_name
						else:
							obj.data.name = new_name
						obj.name = new_name
					
					obj.location = position
					obj.scale = scale
					
					obj.rotation_mode = 'AXIS_ANGLE'
					obj.rotation_axis_angle[0] = -rotation[3]*math.pi/180.0
					obj.rotation_axis_angle[1] = rotation[0]
					obj.rotation_axis_angle[2] = rotation[1]
					obj.rotation_axis_angle[3] = rotation[2]
					
				# deselect all objects
				bpy.ops.object.select_all(action='DESELECT')
			
			elif "new SpawnSphere()" in line:
				radius = lines[i+3].split(" = ")[1].replace(";", "").replace('"', '')
				radius = float(radius)
				
				position = lines[i+9].split(" = ")[1].replace(";", "").replace('"', '')
				position = [float(item) for item in position.split()]
				
				rotation = lines[i+10].split(" = ")[1].replace(";", "").replace('"', '')
				rotation = [float(item) for item in rotation.split()]
				
				scale = lines[i+11].split(" = ")[1].replace(";", "").replace('"', '')
				scale = [float(item) for item in scale.split()]
				
				bpy.ops.object.empty_add(type='SPHERE', radius=radius, location=position)
				spawnsphere_object = bpy.context.scene.objects.active
				spawnsphere_object.name = "SpawnSphere"
				
			elif "new CameraBookmark" in line:
				position = lines[i+3].split(" = ")[1].replace(";", "").replace('"', '')
				position = [float(item) for item in position.split()]
				
				rotation = lines[i+4].split(" = ")[1].replace(";", "").replace('"', '')
				rotation = [float(item) for item in rotation.split()]
				
				scale = lines[i+5].split(" = ")[1].replace(";", "").replace('"', '')
				scale = [float(item) for item in scale.split()]
				
				internalName = lines[i+6].split(" = ")[1].split(";")[0].replace('"', '')
				
				bpy.ops.object.camera_add(location=position)
				camera_object = bpy.context.scene.objects.active
				camera_object.name = internalName
				
				camera_object.scale = scale
				camera_object.rotation_mode = 'AXIS_ANGLE'
				camera_object.rotation_axis_angle[0] = -rotation[3]*math.pi/180.0
				camera_object.rotation_axis_angle[1] = rotation[0]
				camera_object.rotation_axis_angle[2] = rotation[1]
				camera_object.rotation_axis_angle[3] = rotation[2]
			
			elif "new PointLight" in line:
				radius = 1.0
				isEnabled = 1
				color = [1.0,1.0,1.0]
				brightness = 1.0
				flare_scale = 1.0
				position = [0.0, 0.0, 0.0]
				rotation = [0.0, 0.0, 0.0]
				
				for j in range(1, 28):
					parameter, value = lines[i+j].split(" = ")
					parameter = parameter.lstrip()
					value = value.replace(";", "").replace('"', '')
					if parameter == "radius":
						radius = float(value)
					elif parameter == "isEnabled":
						isEnabled = int(value)
					elif parameter == "color":
						color = [float(item) for item in value.split()]
					elif parameter == "brightness":
						brightness = float(value)
					elif parameter == "flare_scale":
						flare_scale = float(value)
					elif parameter == "position":
						position = [float(item) for item in value.split()]
					elif parameter == "rotation":
						rotation = [float(item) for item in value.split()]
				
				bpy.ops.object.lamp_add(type='POINT', radius=radius, location=position)
				pointLight_object = bpy.context.scene.objects.active
				
				pointLight_object.rotation_mode = 'AXIS_ANGLE'
				pointLight_object.rotation_axis_angle[0] = -rotation[3]*math.pi/180.0
				pointLight_object.rotation_axis_angle[1] = rotation[0]
				pointLight_object.rotation_axis_angle[2] = rotation[1]
				pointLight_object.rotation_axis_angle[3] = rotation[2]
				
				pointLight_object.data.color = color[:-1]
				pointLight_object.data.energy = brightness
			
			elif "new SpotLight" in line:
				range_ = 20.0
				innerAngle = 40.0
				outerAngle = 100.0
				isEnabled = 1
				color = [1.0,1.0,1.0]
				brightness = 1.0
				flare_scale = 1.0
				position = [0.0, 0.0, 0.0]
				rotation = [0.0, 0.0, 0.0]
				for j in range(1, 29):
					parameter, value = lines[i+j].split(" = ")
					parameter = parameter.lstrip()
					value = value.replace(";", "").replace('"', '')
					if parameter == "range":
						range_ = float(value)
					elif parameter == "innerAngle":
						innerAngle = float(value)
					elif parameter == "outerAngle":
						outerAngle = float(value)
					elif parameter == "isEnabled":
						isEnabled = int(value)
					elif parameter == "color":
						color = [float(item) for item in value.split()]
					elif parameter == "brightness":
						brightness = float(value)
					elif parameter == "flare_scale":
						flare_scale = float(value)
					elif parameter == "position":
						position = [float(item) for item in value.split()]
					elif parameter == "rotation":
						rotation = [float(item) for item in value.split()]
				
				bpy.ops.object.lamp_add(type='SPOT', radius=range_, location=position)
				spotLight_object = bpy.context.scene.objects.active
				
				spotLight_object.rotation_mode = 'AXIS_ANGLE'
				spotLight_object.rotation_axis_angle[0] = -rotation[3]*math.pi/180.0
				spotLight_object.rotation_axis_angle[1] = rotation[0]
				spotLight_object.rotation_axis_angle[2] = rotation[1]
				spotLight_object.rotation_axis_angle[3] = rotation[2]
				
				spotLight_object.data.color = color[:-1]
				spotLight_object.data.energy = brightness
				spotLight_object.data.distance = range_
	
	MaterialRename()
	DeleteUnesedMaterials()

def MaterialRename():
	mats = bpy.data.materials
	for obj in bpy.data.objects:
		object_name = obj.name
		object_name2 = object_name.rpartition('.')[0]
		if object_name2 != "":
			object_name = object_name2
			
		for mat_slot in obj.material_slots:
			material_name = mat_slot.name
			material_name2 = material_name.rpartition('.')[0]
			if material_name2 != "":
				material_name = material_name2
			
			textures_names = ''
			for mtex_slot in mat_slot.material.texture_slots:
				if mtex_slot:
					if hasattr(mtex_slot.texture , 'image'):
						if (mtex_slot.texture.image is not None) and (mtex_slot.texture.image.size[0] != 0 and mtex_slot.texture.image.size[1] != 0):
							texture_name = mtex_slot.texture.image.name
							texture_name2 = texture_name.rpartition('.')[0]
							if texture_name2 != "":
								texture_name = texture_name2
							textures_names += texture_name
							textures_names += "_"
			
			if textures_names != "":
				textures_names = textures_names[:-1]	# tirando o underline
				material_name = material_name + "_" + textures_names
			else:
				material_name = material_name + "_" + object_name
			
			if material_name in mats:
				mat_slot.material = mats.get(material_name)
			else:
				mat_slot.material.name = material_name
			mats = bpy.data.materials
	
	used_materials = []
	for obj in bpy.data.objects:
		for slot in obj.material_slots:
			if slot.material not in used_materials:
				used_materials.append(slot.material)		

	for material in bpy.data.materials:
		if material not in used_materials:
			material.user_clear()
			bpy.data.materials.remove(material)

def DeleteUnesedMaterials():
	for block in bpy.data.meshes:
		if block.users == 0:
			bpy.data.meshes.remove(block)

	for block in bpy.data.materials:
		if block.users == 0:
			bpy.data.materials.remove(block)

	for block in bpy.data.textures:
		if block.users == 0:
			bpy.data.textures.remove(block)

	for block in bpy.data.images:
		if block.users == 0:
			bpy.data.images.remove(block)
	
	for element in bpy.context.scene.objects:
		if element.type != "MESH": continue
		if bpy.app.version >= (2, 80, 0):
			bpy.context.view_layer.objects.active = element
		else:
			bpy.context.scene.objects.active = element
		mats = element.material_slots
		used_slots = []
		for poly in element.data.polygons:
			if poly.material_index not in used_slots:
				used_slots.append(poly.material_index)
		for i, mat in reversed(list(enumerate((mats)))):
			if i not in used_slots:
				element.active_material_index = i
				bpy.ops.object.material_slot_remove()
	for material in bpy.data.materials:
		if not material.users:
			bpy.data.materials.remove(material)

def ClearScene(context):
	for obj in bpy.context.scene.objects:
		if obj.type == 'MESH' or obj.type == 'EMPTY' or obj.type == 'LAMP':
			if bpy.app.version >= (2, 80, 0):
				obj.select_set(True)
			else:
				obj.select = True
		else:
			if bpy.app.version >= (2, 80, 0):
				obj.select_set(False)
			else:
				obj.select = False
	bpy.ops.object.delete()
	
	for block in bpy.data.meshes:
		if block.users == 0:
			bpy.data.meshes.remove(block)

	for block in bpy.data.materials:
		if block.users == 0:
			bpy.data.materials.remove(block)

	for block in bpy.data.textures:
		if block.users == 0:
			bpy.data.textures.remove(block)

	for block in bpy.data.images:
		if block.users == 0:
			bpy.data.images.remove(block)

def main(context, LevelPath, clear_Scene, edit_texture_paths):
	# Cleaning windown output
	os.system('cls')
	
	print ("--------------------------------------------------")
	print ("---------SCRIPT EXECUTING PYTHON IMPORTER---------")
	print ("--------------------------------------------------")
	
	# Setting blender to object mode
	if bpy.ops.object.mode_set.poll():
		bpy.ops.object.mode_set(mode='OBJECT')
	
	if clear_Scene == True:
		ClearScene(context)
	
	if bpy.app.version >= (2, 80, 0):
		bpy.context.scene.cursor.location = (0.0, 0.0, 0.0)
	else:
		bpy.context.scene.cursor_location = (0.0, 0.0, 0.0)
	
	# .mis path
	misPath = os.path.join(LevelPath, "GridMap_v0422.mis")
	
	# Read .mis
	read_mis(misPath, edit_texture_paths)
	
	if bpy.context.scene.world is None:
		newworld = bpy.data.worlds.new("World")
		bpy.context.scene.world = newworld
	bpy.context.scene.world.name = 'Level'
	
	for area in bpy.context.screen.areas:
		if area.type == 'VIEW_3D':
			region = next(region for region in area.regions if region.type == 'WINDOW')
			override = bpy.context.copy()
			override['area'] = area
			override['region'] = region
			bpy.ops.view3d.view_all(override, use_all_regions=False, center=False)
			area.spaces.active.clip_end = 100000
			space = next(space for space in area.spaces if space.type == 'VIEW_3D')
			if bpy.app.version >= (2, 80, 0):
				space.shading.type = 'MATERIAL'	# set the viewport shading to rendered
			else:
				space.viewport_shade = 'MATERIAL' # set the viewport shading to rendered
			break

	# Create lamp
	if bpy.app.version >= (2, 80, 0):
		bpy.ops.object.light_add(type='SUN',location=(0,0,3))
		bpy.context.active_object.name = 'Sun'
		bpy.context.active_object.hide_viewport = True
	else:
		bpy.ops.object.lamp_add(type='HEMI',location=(0,0,3))
		bpy.context.active_object.name = 'Sun'
		bpy.context.active_object.hide = True
	
	if bpy.app.version >= (2, 80, 0):
		bpy.context.view_layer.objects.active = None
	else:
		bpy.context.scene.objects.active = None
	bpy.ops.object.select_all(action='DESELECT')
	
	return {'FINISHED'}

class ImportBeamNG_driveLevels(Operator, ImportHelper):
	"""Load a BeamNG.drive level"""
	bl_idname = "import_beamnglevel.data"  # important since its how bpy.ops.import_test.some_data is constructed
	bl_label = "Import level"

	# List of operator properties, the attributes will be assigned
	# to the class instance from the operator settings before calling.
	clear_Scene = BoolProperty(
			name="Clear scene",
			description="Check in order to clear the scene",
			default=True,
			)
	
	edit_texture_paths = BoolProperty(
			name="Edit texture paths",
			description="Check in order to replace the texture paths in the .dae files to the texture name",
			default=True,
			)
	
	def execute(self, context):
		return main(context, self.filepath, self.clear_Scene, self.edit_texture_paths)

def make_annotations(cls):
	"""Converts class fields to annotations if running with Blender 2.8"""
	if bpy.app.version < (2, 80):
		return cls
	bl_props = {k: v for k, v in cls.__dict__.items() if isinstance(v, tuple)}
	if bl_props:
		if '__annotations__' not in cls.__dict__:
			setattr(cls, '__annotations__', {})
		annotations = cls.__dict__['__annotations__']
		for k, v in bl_props.items():
			annotations[k] = v
			delattr(cls, k)
	return cls

# Only needed if you want to add into a dynamic menu
def menu_func_import(self, context):
	self.layout.operator(ImportBeamNG_driveLevels.bl_idname, text="BeamNG.drive Level")


def register():
	if bpy.app.version >= (2, 80, 0):
		make_annotations(ImportBeamNG_driveLevels)
		bpy.utils.register_class(ImportBeamNG_driveLevels)
		bpy.types.TOPBAR_MT_file_import.append(menu_func_import)
	else:
		bpy.utils.register_class(ImportBeamNG_driveLevels)
		bpy.types.INFO_MT_file_import.append(menu_func_import)

def unregister():
	if bpy.app.version >= (2, 80, 0):
		bpy.types.TOPBAR_MT_file_import.remove(menu_func_import)
	else:
		bpy.types.INFO_MT_file_import.remove(menu_func_import)
	bpy.utils.unregister_class(ImportBeamNG_driveLevels)

if __name__ == "__main__":
	register()
	
	# test call
	bpy.ops.import_beamnglevel.data('INVOKE_DEFAULT')